/*******************************************************************************
 * Copyright  (c) 2011, 2012, 2013 University of Pennsylvania
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without a
 * written agreement is hereby granted, provided that the above copyright
 * notice and this paragraph and the following two paragraphs appear in
 * all copies.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
 * AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*******************************************************************************/

//Parametrized Aries Fat Pointer Implementation

import Types :: *;

interface AriesIfc#(numeric type a_len,
                    numeric type b_len,
                    numeric type l_len);
        method Tuple2#(Aries_FatPointer#(a_len, b_len, l_len), Bool)
                addp (Aries_FatPointer#(a_len, b_len, l_len) ptr,
                      Int#(a_len) offset);
        method Int#(TAdd#(a_len, 1))
                offsetp (Aries_FatPointer#(a_len, b_len, l_len) ptr);
endinterface

module mkAries (AriesIfc#(a_len, b_len, l_len))
        provisos (Add#(a__, l_len, a_len));

        function Bit#(a_len) basep (Aries_FatPointer#(a_len, b_len, l_len) ptr);
                Bit#(a_len) lminf = extend(ptr.l - ptr.f);
                Bit#(a_len) twotob = 1 << ptr.b;
                Bit#(a_len) base =
                    ptr.a - (ptr.a & (twotob - 1)) - (extend(ptr.f) << ptr.b);
                return base;
        endfunction

        //returns updated pointer and is within bounds
        //bounds check on an existing pointer is essentially the case when
        //offset is 0
        method Tuple2#(Aries_FatPointer#(a_len, b_len, l_len), Bool)
                addp (Aries_FatPointer#(a_len, b_len, l_len) ptr,
                      Int#(a_len) offset);

                Bit#(a_len) lminf = extend(ptr.l - ptr.f);
                Bit#(a_len) twotob = 1 << ptr.b;
                Bit#(a_len) base =
                    ptr.a - (ptr.a & (twotob - 1)) - (extend(ptr.f) << ptr.b);

                Int#(TAdd#(a_len, 1)) tmp_dest =
                    unpack(zeroExtend(ptr.a)) + extend(offset);

                Bit#(a_len) dest = truncate(pack( tmp_dest ));
                Bit#(a_len) bound =
                    (ptr.a - (ptr.a & (twotob-1))) + ((lminf + 1) << ptr.b);
                Bit#(l_len) new_f =
                    truncate( (dest - (dest & (twotob -1)) - base) >> ptr.b );

                if ((dest > bound) || (dest < base)) begin
                        //pointer went out of bounds
                        Aries_FatPointer#(a_len, b_len, l_len) new_ptr =
                            Aries_FatPointer {g: ERROR,
                                              l: 0, b: 0, f: 0, a: 0};
                        return tuple2(new_ptr, False);
                end
                else begin
                        //successful bounds check
                        Aries_FatPointer#(a_len, b_len, l_len) new_ptr =
                            Aries_FatPointer {g: ptr.g,
                                              l: ptr.l, b: ptr.b,
                                              f: new_f, a: dest};
                        return tuple2(new_ptr, True);
                end

        endmethod

        method Int#(TAdd#(a_len, 1))
                offsetp (Aries_FatPointer#(a_len, b_len, l_len) ptr);

                Bit#(a_len) base = basep (ptr);
                return unpack(extend( ptr.a - base ));

        endmethod

endmodule

(* synthesize *)
module mkAriesSynth (AriesIfc#(`A, `B, `L));
        AriesIfc#(`A, `B, `L) _a <- mkAries;
        return _a;
endmodule
