/*******************************************************************************
 * Copyright  (c) 2011, 2012, 2013 University of Pennsylvania
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without a
 * written agreement is hereby granted, provided that the above copyright
 * notice and this paragraph and the following two paragraphs appear in
 * all copies.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
 * AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 ******************************************************************************/

//Parametrized BIMA Fat Pointer Implementation

import Types :: *;

interface BIMAIfc#(numeric type a_len, numeric type b_len, numeric type l_len);
        //returns the distance from base pointer
        method Int#(a_len) offsetp (BIMA_FatPointer#(a_len, b_len, l_len) ptr);
        //returns new resulting pointer, and pointer in bounds or not
        method Tuple2#(BIMA_FatPointer#(a_len, b_len, l_len), Bool)
                addp (BIMA_FatPointer#(a_len, b_len, l_len) ptr,
                      Int#(a_len) offset);
        //decodes dunder/dover based on the bima given
        method BIMA_FatPointer#(a_len, b_len, l_len)
                decode (BIMA_FatPointer#(a_len, b_len, l_len) ptr);
endinterface

module mkBIMA (BIMAIfc#(a_len, b_len, l_len))
        provisos (Add#(a__, l_len, a_len));

        method offsetp (ptr) = ptr.dunder;

        method Tuple2#(BIMA_FatPointer#(a_len, b_len, l_len), Bool)
                addp (BIMA_FatPointer#(a_len, b_len, l_len) ptr,
                      Int#(a_len) offset);

                Int#(a_len) tmp_dunder = ptr.dunder + offset;
                Int#(a_len) tmp_dover = ptr.dover - offset;
                Int#(TAdd#(a_len, 1)) tmp_dest =
                    unpack(zeroExtend(ptr.a)) + extend(offset);
                Bit#(a_len) dest = truncate(pack( tmp_dest ));
                if (tmp_dunder < 0 || tmp_dover < 0)
                        return tuple2(BIMA_FatPointer {g: ERROR,
                                                       b: 0, i: 0, m: 0, a: 0,
                                                       dunder: tmp_dunder,
                                                       dover:tmp_dover},
                                      False);
                else
                        return tuple2(BIMA_FatPointer {g: ptr.g,
                                                       b: ptr.b, i: ptr.i,
                                                       m: ptr.m, a: dest,
                                                       dunder:tmp_dunder,
                                                       dover:tmp_dover},
                                      True);

        endmethod

        method BIMA_FatPointer#(a_len, b_len, l_len)
                decode (BIMA_FatPointer#(a_len, b_len, l_len) ptr);

                Bit#(a_len) dunder_1;
                Bit#(a_len) dover_1;
                Bit#(a_len) dunder_2;
                Bit#(a_len) dover_2;

                Bit#(a_len) carry = (64 << ptr.b);
                Bit#(a_len) ones = extend(carry - 1);

                Bit#(a_len) a_aboveb = ptr.a >> ptr.b;
                Bit#(a_len) i_shiftb = extend(ptr.i) << ptr.b;
                Bit#(a_len) m_shiftb = extend(ptr.m) << ptr.b;

                dunder_1 = (carry  | (ones & ptr.a)) - i_shiftb;
                dunder_2 = (ones & ptr.a) - i_shiftb;

                dover_1 = (carry | m_shiftb) - (ones & ptr.a) - 1;
                dover_2 = m_shiftb - (ones & ptr.a) - 1;

                return BIMA_FatPointer {g:ptr.g,
                                        b: ptr.b, i: ptr.i, m: ptr.m, a: ptr.a,
                                        dunder: (ptr.i > a_aboveb[valueof(l_len)-1:0] ?
                                                 unpack(dunder_1) :
                                                 unpack(dunder_2)),
                                        dover:  (ptr.m < a_aboveb[valueof(l_len)-1:0] ?
                                                 unpack(dover_1) :
                                                 unpack(dover_2))};
        endmethod

endmodule

(* synthesize *)
module mkBIMASynth (BIMAIfc#(`A, `B, `L));
        BIMAIfc#(`A, `B, `L) _a <- mkBIMA;
        return _a;
endmodule
